import logging
import shutil
from pathlib import Path

import mutagen

import audio_scan_info

SAMPLES_PATH = Path("tests", "sample_files").resolve()
LOG_LEVEL = logging.INFO

EXPECTED_ARTIST = "Test Artist"
EXPECTED_TITLE = "Test Title"
EXPECTED_LENGTH = "00:01"
EXPECTED_SAMPLE_RATE = "8.0 kHz"
EXPECTED_CHANNELS = "1"
EXPECTED_MP3_BITRATE = "8 kbps"
EXPECTED_MP3_BITRATE_MODE = "CBR"
EXPECTED_FLAC_BITRATE = "45 kbps"
EXPECTED_FLAC_BITS = "16"


def copy_file(filename, path):
    samples_filepath = Path("tests", "sample_files", filename).resolve()
    Path(path / Path(filename).parent).mkdir(parents=True, exist_ok=True)
    shutil.copyfile(SAMPLES_PATH / filename, path / filename)
    try:
        audio = mutagen.File(samples_filepath, easy=True)
        assert EXPECTED_ARTIST in audio["artist"], "test files are tagged incorrectly"
        assert EXPECTED_TITLE in audio["title"], "test files are tagged incorrectly"
    except mutagen.MutagenError:
        pass


def test_get_metadata_mp3(tmp_path):
    filename = "test_file.mp3"
    copy_file(filename, tmp_path)
    filepath = tmp_path / filename
    tags, info = audio_scan_info.get_metadata(filepath)
    assert len(tags) == 2
    assert tags["artist"] == EXPECTED_ARTIST
    assert tags["title"] == EXPECTED_TITLE
    assert len(info) == 5
    assert info["length"] == EXPECTED_LENGTH
    assert info["bitrate"] == EXPECTED_MP3_BITRATE
    assert info["bitrate mode"] == EXPECTED_MP3_BITRATE_MODE
    assert info["sample rate"] == EXPECTED_SAMPLE_RATE
    assert info["channels"] == EXPECTED_CHANNELS


def test_get_metadata_flac(tmp_path):
    filename = "test_file.flac"
    copy_file(filename, tmp_path)
    filepath = tmp_path / filename
    tags, info = audio_scan_info.get_metadata(filepath)
    assert len(tags) == 2
    assert tags["artist"] == EXPECTED_ARTIST
    assert tags["title"] == EXPECTED_TITLE
    assert len(info) == 5
    assert info["length"] == EXPECTED_LENGTH
    assert info["bitrate"] == EXPECTED_FLAC_BITRATE
    assert info["bits per sample"] == EXPECTED_FLAC_BITS
    assert info["sample rate"] == EXPECTED_SAMPLE_RATE
    assert info["channels"] == EXPECTED_CHANNELS


def test_analyze_mp3(tmp_path, caplog):
    filename = "test_file.mp3"
    copy_file(filename, tmp_path)
    filepath = tmp_path / filename
    caplog.set_level(LOG_LEVEL)
    audio_scan_info.analyze_file(filepath)
    for record in caplog.records:
        assert record.levelname == "INFO"
    assert f"\u27a4  File: {filepath}\n" in caplog.text
    assert "   \u2794 Info:\n" in caplog.text
    assert f"       length: {EXPECTED_LENGTH}\n" in caplog.text
    assert f"       bitrate: {EXPECTED_MP3_BITRATE}\n" in caplog.text
    assert f"       sample rate: {EXPECTED_SAMPLE_RATE}\n" in caplog.text
    assert f"       bitrate mode: {EXPECTED_MP3_BITRATE_MODE}\n" in caplog.text
    assert f"       channels: {EXPECTED_CHANNELS}\n" in caplog.text
    assert "   \u2794 Tags:\n" in caplog.text
    assert f"       artist: {EXPECTED_ARTIST}\n" in caplog.text
    assert f"       title: {EXPECTED_TITLE}\n" in caplog.text


def test_analyze_flac(tmp_path, caplog):
    filename = "test_file.flac"
    copy_file(filename, tmp_path)
    filepath = tmp_path / filename
    caplog.set_level(LOG_LEVEL)
    audio_scan_info.analyze_file(filepath)
    for record in caplog.records:
        assert record.levelname == "INFO"
    assert f"\u27a4  File: {filepath}\n" in caplog.text
    assert "   \u2794 Info:\n" in caplog.text
    assert f"       length: {EXPECTED_LENGTH}\n" in caplog.text
    assert f"       bitrate: {EXPECTED_FLAC_BITRATE}\n" in caplog.text
    assert f"       bits per sample: {EXPECTED_FLAC_BITS}\n" in caplog.text
    assert f"       sample rate: {EXPECTED_SAMPLE_RATE}\n" in caplog.text
    assert f"       channels: {EXPECTED_CHANNELS}\n" in caplog.text
    assert "   \u2794 Tags:\n" in caplog.text
    assert f"       artist: {EXPECTED_ARTIST}\n" in caplog.text
    assert f"       title: {EXPECTED_TITLE}\n" in caplog.text


def test_analyze_unknown_file(caplog):
    filename = Path("unknown_file.mp3")
    caplog.set_level(LOG_LEVEL)
    audio_scan_info.analyze_file(filename)
    for record in caplog.records:
        assert record.levelname == "ERROR"
    assert f"\u27a4  File: {filename}\n"
    assert "   \u2717 Error:\n     can't find file\n" in caplog.text


def test_run_filenames(tmp_path, caplog):
    filenames = [f"test_file.{extension}" for extension in ("mp3", "flac")]
    assert len(filenames) == 2
    for filename in filenames:
        copy_file(filename, tmp_path)
    caplog.set_level(LOG_LEVEL)
    audio_scan_info.run(tmp_path, filenames)
    for record in caplog.records:
        assert record.levelname == "INFO"
    for filepath in (tmp_path / filename for filename in filenames):
        assert f"\u27a4  File: {filepath}\n"
    assert "   \u2794 Info:\n" in caplog.text
    assert f"       length: {EXPECTED_LENGTH}\n" in caplog.text
    assert f"       bitrate: {EXPECTED_MP3_BITRATE}\n" in caplog.text
    assert f"       sample rate: {EXPECTED_SAMPLE_RATE}\n" in caplog.text
    assert f"       bitrate mode: {EXPECTED_MP3_BITRATE_MODE}\n" in caplog.text
    assert f"       channels: {EXPECTED_CHANNELS}\n" in caplog.text
    assert "   \u2794 Info:\n" in caplog.text
    assert f"       length: {EXPECTED_LENGTH}\n" in caplog.text
    assert f"       bitrate: {EXPECTED_FLAC_BITRATE}\n" in caplog.text
    assert f"       bits per sample: {EXPECTED_FLAC_BITS}\n" in caplog.text
    assert f"       sample rate: {EXPECTED_SAMPLE_RATE}\n" in caplog.text
    assert f"       channels: {EXPECTED_CHANNELS}\n" in caplog.text
    assert "   \u2794 Tags:\n" in caplog.text
    assert f"       artist: {EXPECTED_ARTIST}\n" in caplog.text
    assert f"       title: {EXPECTED_TITLE}\n" in caplog.text


def test_run_dir(tmp_path, caplog):
    filenames = [path.relative_to(SAMPLES_PATH) for path in SAMPLES_PATH.rglob("*") if path.is_file()]
    num_files = len(filenames)
    for filename in filenames:
        copy_file(filename, tmp_path)
    caplog.set_level(LOG_LEVEL)
    audio_scan_info.run(tmp_path, [])
    for record in caplog.records:
        assert record.levelname in ("ERROR", "INFO")
    temp_files = [tmp_path / filename for filename in filenames]
    assert num_files == len(temp_files)
    for filepath in temp_files:
        assert f"\u27a4  File: {filepath}\n" in caplog.text



# Copyright (c) 2025 Corey Goldberg
# MIT License

import argparse
import logging
import sys
import time
from pathlib import Path

import mutagen
import mutagen.apev2
import mutagen.easyid3

logging.basicConfig(level=logging.INFO, stream=sys.stdout, format="%(message)s")
logger = logging.getLogger(__name__)


SUPPORTED_FORMATS = ("FLAC", "MP3")
FILE_EXTENSIONS = tuple(f".{x.lower()}" for x in SUPPORTED_FORMATS)


def get_metadata(filepath):
    file_label = f"{light_blue_arrowhead()}  File: {filepath}"
    info = {}
    tags = {}
    try:
        audio = mutagen.File(filepath, easy=True)
        if audio is None:
            logger.error(f"{file_label}\n   {red_x()} Error:\n     unknown error\n")
        else:
            if "audio/x-mp3" in audio.mime:
                info["length"] = time.strftime("%M:%S", time.gmtime(audio.info.length))
                info["bitrate"] = f"{int(audio.info.bitrate / 1000)} kbps"
                info["sample rate"] = f"{(audio.info.sample_rate / 1000.0):.1f} kHz"
                info["bitrate mode"] = str(audio.info.bitrate_mode).split(".")[-1]
                info["channels"] = str(audio.info.channels)
                if audio.info.track_gain:
                    info["track gain"] = f"{audio.info.track_gain:.1f} db"
                if audio.info.track_peak:
                    info["track peak"] = f"{audio.info.track_peak:.1f}"
                if audio.info.album_gain:
                    info["album gain"] = f"{audio.info.album_gain:.1f} db"
                id3_tags = {tag: value[0] for tag, value in audio.tags.items()} if audio.tags is not None else {}
                # some MP3's have APEv2 tags also
                ape = mutagen.apev2.APEv2File(filepath)
                ape_tags = {}
                if ape.tags is not None:
                    for tag, value in ape.tags.items():
                        ape_tags[f"ape_{tag.lower()}"] = (
                            "<BINARY>" if isinstance(value, mutagen.apev2.APEBinaryValue) else value
                        )
                tags = id3_tags | ape_tags
            elif "audio/x-flac" in audio.mime:
                info["length"] = time.strftime("%M:%S", time.gmtime(audio.info.length))
                info["bitrate"] = f"{int(audio.info.bitrate / 1000)} kbps"
                info["bits per sample"] = str(audio.info.bits_per_sample)
                info["sample rate"] = f"{(audio.info.sample_rate / 1000.0):.1f} kHz"
                info["channels"] = str(audio.info.channels)
                if audio.pictures:
                    info["has embedded pictures"] = "true"
                if audio.cuesheet:
                    info["has cuesheet"] = "true"
                if audio.seektable:
                    info["has seek table"] = "true"
                if audio.tags is None:
                    tags = {}
                else:
                    tags = {tag: value for tag, value in audio.tags}
    except Exception as e:
        logger.error(f"{file_label}\n   {red_x()} Error:\n     {e}\n")
    return tags, info


def analyze_file(filepath):
    file_label = f"{light_blue_arrowhead()}  File: {filepath}"
    if not filepath.exists():
        logger.error(f"{file_label}\n   {red_x()} Error:\n     can't find file\n")
        return
    if filepath.name.lower().endswith(FILE_EXTENSIONS):
        tags, info = get_metadata(filepath)
        if info:
            logger.info(f"{file_label}\n   {light_blue_arrow()} Info:")
            for attr, value in info.items():
                logger.info(f"       {attr}: {value}")
        if tags:
            logger.info(f"   {light_blue_arrow()} Tags:")
            for tag, value in tags.items():
                logger.info(f"       {tag}: {value}")
        logger.info("")


def run(path, filenames):
    if filenames:
        for f in filenames:
            filepath = Path(path / f).resolve()
            analyze_file(filepath)
    else:
        for root, dirs, files in path.walk():
            for f in sorted(files):
                filepath = Path(root / f).resolve()
                analyze_file(filepath)


def colored_symbol(unicode_char, rgb):
    r, g, b = rgb
    return f"\033[38;2;{r};{g};{b}m{unicode_char}\033[0m"


def red_x():
    return colored_symbol("\u2717", (255, 0, 0))


def light_blue_arrow():
    return colored_symbol("\u2794", (144, 213, 255))


def light_blue_arrowhead():
    return colored_symbol("\u27a4", (144, 213, 255))


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("filename", nargs="*", help="[optional] file to process (multiple allowed)")
    parser.add_argument("-d", "--dir", default=Path.cwd().resolve(), help="start directory")
    args = parser.parse_args()
    path = Path(args.dir)
    filenames = sorted(Path(f) for f in args.filename)
    if not path.exists():
        sys.exit(f"{red_x()} Error: can't find '{path}'")
    try:
        run(path, filenames)
    except KeyboardInterrupt:
        sys.exit()
